package ru.dbotthepony.kstarbound.server.world

import com.github.benmanes.caffeine.cache.Cache
import com.github.benmanes.caffeine.cache.Caffeine
import com.github.benmanes.caffeine.cache.Scheduler
import com.google.gson.stream.JsonReader
import it.unimi.dsi.fastutil.io.FastByteArrayInputStream
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import ru.dbotthepony.kommons.collect.chainOptionalFutures
import ru.dbotthepony.kommons.gson.get
import ru.dbotthepony.kommons.io.BTreeDB6
import ru.dbotthepony.kommons.util.AABBi
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.CelestialBaseInformation
import ru.dbotthepony.kstarbound.defs.CelestialGenerationInformation
import ru.dbotthepony.kstarbound.defs.CelestialNames
import ru.dbotthepony.kstarbound.fromJson
import ru.dbotthepony.kstarbound.io.BTreeDB5
import ru.dbotthepony.kstarbound.util.random.staticRandom64
import ru.dbotthepony.kstarbound.world.CoordinateMapper
import ru.dbotthepony.kstarbound.world.Universe
import ru.dbotthepony.kstarbound.world.UniversePos
import ru.dbotthepony.kstarbound.world.positiveModulo
import java.io.Closeable
import java.io.File
import java.io.InputStreamReader
import java.time.Duration
import java.util.*
import java.util.concurrent.CompletableFuture
import kotlin.collections.ArrayList

class ServerUniverse private constructor(marker: Nothing?) : Universe(), Closeable {
	constructor() : this(null) {
		sources.add(NativeUniverseSource(null, this).generator)
	}

	constructor(folder: File) : this(null) {
		val nativeFile = File(folder, "universe.kchunks")
		val legacyFile = File(folder, "universe.chunks")

		val native = if (!nativeFile.exists()) {
			NativeUniverseSource(BTreeDB6.create(nativeFile, sync = false), this)
		} else {
			NativeUniverseSource(BTreeDB6(nativeFile, sync = false), this)
		}

		val legacy = if (legacyFile.exists()) {
			LegacyUniverseSource(BTreeDB5(legacyFile))
		} else {
			null
		}

		sources.add(native.reader)
		if (legacy != null) sources.add(legacy)
		sources.add(native.generator)

		closeables.add(native)
		if (legacy != null) closeables.add(legacy)
	}

	override val baseInformation: CelestialBaseInformation
	val generationInformation: CelestialGenerationInformation
	val celestialNames: CelestialNames

	private val sources = ArrayList<UniverseSource>()
	private val closeables = ArrayList<Closeable>()

	override fun name(pos: UniversePos): CompletableFuture<KOptional<String>> {
		return getChunk(pos).thenApply {
			it.flatMap { it.parameters(pos) }.map { it.name }
		}
	}

	override fun scanSystems(region: AABBi, includedTypes: Set<String>?): CompletableFuture<List<UniversePos>> {
		val copy = if (includedTypes != null) ObjectOpenHashSet(includedTypes) else null
		val futures = ArrayList<CompletableFuture<List<UniversePos>>>()

		for (pos in chunkPositions(region)) {
			val f = getChunk(pos).thenApply {
				it.map<List<UniversePos>> {
					val result = ArrayList<UniversePos>()

					if (copy == null) {
						for (system in it.systems.keys) {
							result.add(UniversePos(system))
						}
					} else {
						for ((system, params) in it.systems) {
							if (params.parameters.parameters.get("typeName", "") in copy) {
								result.add(UniversePos(system))
							}
						}
					}

					result
				}.orElse(listOf())
			}

			futures.add(f)
		}

		return CompletableFuture.allOf(*futures.toTypedArray())
			.thenApply { futures.stream().flatMap { it.get().stream() }.toList() }
	}

	override fun scanConstellationLines(region: AABBi): CompletableFuture<List<Pair<Vector2i, Vector2i>>> {
		val futures = ArrayList<CompletableFuture<List<Pair<Vector2i, Vector2i>>>>()

		for (pos in chunkPositions(region)) {
			val f = getChunk(pos).thenApply {
				it.map<List<Pair<Vector2i, Vector2i>>> { ObjectArrayList(it.constellations) }.orElse(listOf())
			}

			futures.add(f)
		}

		return CompletableFuture.allOf(*futures.toTypedArray())
			.thenApply { futures.stream().flatMap { it.get().stream() }.toList() }
	}

	override fun scanRegionFullyLoaded(region: AABBi): Boolean {
		return true
	}

	override fun close() {
		closeables.forEach { it.close() }
	}

	// Edge case: we load so many chunks AND try to load same chunk twice that it gets loaded/generated twice
	// shouldn't cause actual issues though
	private val chunkCache: Cache<Vector2i, CompletableFuture<KOptional<UniverseChunk>>> = Caffeine
		.newBuilder()
		.expireAfterAccess(Duration.ofMinutes(10L))
		.maximumSize(1024L)
		.softValues()
		.scheduler(Scheduler.systemScheduler())
		.build()

	fun getChunk(pos: UniversePos): CompletableFuture<KOptional<UniverseChunk>> {
		return getChunk(world2chunk(Vector2i(pos.location)))
	}

	fun getChunk(pos: Vector2i): CompletableFuture<KOptional<UniverseChunk>> {
		return chunkCache.get(pos) { p -> chainOptionalFutures(sources) { it.getChunk(p) } }
	}

	init {
		val celestial = Starbound.read("/celestial.config")
		val stream0 = JsonReader(InputStreamReader(FastByteArrayInputStream(celestial.array())))
		val stream1 = JsonReader(InputStreamReader(FastByteArrayInputStream(celestial.array())))

		stream0.isLenient = true
		stream1.isLenient = true

		baseInformation = Starbound.gson.fromJson(stream0)!!
		generationInformation = Starbound.gson.fromJson(stream1)!!

		if (!generationInformation.systemTypePerlin.isInitialized)
			generationInformation.systemTypePerlin.init(staticRandom64("SystemTypePerlin"))

		celestialNames = Starbound.gson.fromJson(Starbound.jsonReader("/celestial/names.config"))!!
	}

	override val region: AABBi = AABBi(Vector2i(baseInformation.xyCoordRange.x, baseInformation.xyCoordRange.x), Vector2i(baseInformation.xyCoordRange.y, baseInformation.xyCoordRange.y))

}
