package ru.dbotthepony.kstarbound.world

import com.github.benmanes.caffeine.cache.Caffeine
import ru.dbotthepony.kommons.util.AABBi
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.defs.CelestialBaseInformation
import java.time.Duration
import java.util.concurrent.CompletableFuture

abstract class Universe {
	abstract val region: AABBi
	abstract val baseInformation: CelestialBaseInformation

	abstract fun name(pos: UniversePos): CompletableFuture<KOptional<String>>

	/**
	 * Return all valid system coordinates in the given x/y range.  All systems
	 * are guaranteed to have unique x/y coordinates, and are meant to be viewed
	 * from the top in 2d.  The z-coordinate is there simpy as a validation
	 * parameter.
	 */
	abstract fun scanSystems(region: AABBi, includedTypes: Set<String>? = null): CompletableFuture<List<UniversePos>>
	abstract fun scanConstellationLines(region: AABBi): CompletableFuture<List<Pair<Vector2i, Vector2i>>>

	/**
	 * Returns false if part or all of the specified region is not loaded.  This
	 * is only relevant for calls to scanSystems and scanConstellationLines, and
	 * does not imply that each individual system in the given region is fully
	 * loaded with all planets moons etc, only that scanSystem and
	 * scanConstellationLines are not waiting on missing data.
	 */
	abstract fun scanRegionFullyLoaded(region: AABBi): Boolean

	fun chunkPositions(region: AABBi): List<Vector2i> {
		if (region.mins == region.maxs)
			return emptyList()

		val result = ArrayList<Vector2i>()

		val mins = world2chunk(region.mins)
		val maxs = world2chunk(region.maxs - Vector2i.POSITIVE_XY)

		for (x in mins.x .. maxs.x) {
			for (y in mins.y .. maxs.y) {
				result.add(Vector2i(x, y))
			}
		}

		return result
	}

	fun world2chunk(pos: IStruct2i): Vector2i {
		val (x, y) = pos

		return Vector2i(
			(x - positiveModulo(x, baseInformation.chunkSize)) / baseInformation.chunkSize,
			(y - positiveModulo(y, baseInformation.chunkSize)) / baseInformation.chunkSize)
	}

	fun chunkRegion(chunkPos: IStruct2i): AABBi {
		val mins = Vector2i(chunkPos) * baseInformation.chunkSize
		return AABBi(mins, mins + Vector2i.POSITIVE_XY * baseInformation.chunkSize)
	}
}
